/**
 * crypto.c
 *
 * Copyright (C) 2006 Bojan Smojver, Rexursive
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "private.h"

/*
 * MD5 crypto functions
 */
static char *digasc(apr_pool_t *pool,unsigned char *digest){
  unsigned char *ascdig=apr_palloc(pool,APR_MD5_DIGESTSIZE*2+1),*d,*a;
  size_t i;

  for(i=0,d=digest,a=ascdig;i<APR_MD5_DIGESTSIZE;i++,d++,a++){
    *a='a'+((*d)/16);
    *(++a)='a'+((*d)%16);
  }
  *a='\0';

  return (char*)ascdig;
}

char *rxv_spin_hash(apr_pool_t *pool,const char *uniq){
  unsigned char *digest=apr_palloc(pool,APR_MD5_DIGESTSIZE+1);

  if(apr_md5(digest,uniq,strlen(uniq))!=APR_SUCCESS)
    return NULL;

  *(digest+APR_MD5_DIGESTSIZE)='\0';

  return digasc(pool,digest);
}

char *rxv_spin_hmac(apr_pool_t *pool,const char *uniq,const char *salt){
  unsigned char *digest=apr_palloc(pool,APR_MD5_DIGESTSIZE+1),
                *ipad=apr_palloc(pool,RXV_SPIN_SALT_MAX+1),
                *opad=apr_palloc(pool,RXV_SPIN_SALT_MAX+1);
  apr_md5_ctx_t *ctx=apr_palloc(pool,sizeof(*ctx));
  size_t i;

  /* we know that our salt (key) is always 64 bytes */
  memcpy(ipad,salt,RXV_SPIN_SALT_MAX);
  memcpy(opad,salt,RXV_SPIN_SALT_MAX);

  *(ipad+RXV_SPIN_SALT_MAX)='\0';
  *(opad+RXV_SPIN_SALT_MAX)='\0';

  for(i=0;i<RXV_SPIN_SALT_MAX;i++){
    ipad[i]^=0x36;
    opad[i]^=0x5c;
  }

  /* round one: using inner pad */
  if(apr_md5_init(ctx)!=APR_SUCCESS ||
     apr_md5_update(ctx,ipad,RXV_SPIN_SALT_MAX)!=APR_SUCCESS ||
     apr_md5_update(ctx,uniq,strlen(uniq))!=APR_SUCCESS ||
     apr_md5_final(digest,ctx)!=APR_SUCCESS)
    return NULL;

  /* round two: using outer pad */
  if(apr_md5_init(ctx)!=APR_SUCCESS ||
     apr_md5_update(ctx,opad,RXV_SPIN_SALT_MAX)!=APR_SUCCESS ||
     apr_md5_update(ctx,digest,APR_MD5_DIGESTSIZE)!=APR_SUCCESS ||
     apr_md5_final(digest,ctx)!=APR_SUCCESS)
    return NULL;

  *(digest+APR_MD5_DIGESTSIZE)='\0';

  return digasc(pool,digest);
}
